@@ -13,7 +13,7 @@ use super::*;
 ///
 /// // The backing array type can be specified in the macro call
 /// let empty_av = array_vec!([u8; 16]);
-/// let some_ints = array_vec!([i32; 4], 1, 2, 3);
+/// let some_ints = array_vec!([i32; 4] => 1, 2, 3);
 ///
 /// // Or left to inference
 /// let empty_av: ArrayVec<[u8; 10]> = array_vec!();
@@ -21,24 +21,21 @@ use super::*;
 /// ```
 #[macro_export]
 macro_rules! array_vec {
-  ($array_type:ty) => {
-    {
-      let av: $crate::ArrayVec<$array_type> = Default::default();
-      av
-    }
-  };
-  ($array_type:ty, $($elem:expr),*) => {
+  ($array_type:ty => $($elem:expr),* $(,)?) => {
     {
       let mut av: $crate::ArrayVec<$array_type> = Default::default();
       $( av.push($elem); )*
       av
     }
   };
-  () => {
-    array_vec!(_)
+  ($array_type:ty) => {
+    $crate::array_vec!($array_type =>)
   };
   ($($elem:expr),*) => {
-    array_vec!(_, $($elem),*)
+    $crate::array_vec!(_ => $($elem),*)
+  };
+  () => {
+    $crate::array_vec!(_)
   };
 }
 
@@ -63,7 +60,7 @@ macro_rules! array_vec {
 /// to have.
 /// ```rust
 /// # use tinyvec::*;
-/// let some_ints = array_vec!([i32; 4], 1, 2, 3);
+/// let some_ints = array_vec!([i32; 4] => 1, 2, 3);
 /// assert_eq!(some_ints.len(), 3);
 /// ```
 ///
@@ -141,19 +138,60 @@ impl<A: Array> ArrayVec<A> {
   /// ## Example
   /// ```rust
   /// # use tinyvec::*;
-  /// let mut av = array_vec!([i32; 10], 1, 2, 3);
-  /// let mut av2 = array_vec!([i32; 10], 4, 5, 6);
+  /// let mut av = array_vec!([i32; 10] => 1, 2, 3);
+  /// let mut av2 = array_vec!([i32; 10] => 4, 5, 6);
   /// av.append(&mut av2);
   /// assert_eq!(av, &[1, 2, 3, 4, 5, 6][..]);
   /// assert_eq!(av2, &[][..]);
   /// ```
   #[inline]
   pub fn append(&mut self, other: &mut Self) {
+    let new_len = self.len() + other.len();
+    assert!(
+      new_len <= A::CAPACITY,
+      "ArrayVec::append> total length {} exceeds capacity {}!",
+      new_len,
+      A::CAPACITY
+    );
+
     for item in other.drain(..) {
-      self.push(item)
+      self.push(item);
     }
   }
 
+  /// Move all values from `other` into this vec.
+  /// If appending would overflow the capacity, Some(other) is returned.
+  /// ## Example
+  /// ```rust
+  /// # use tinyvec::*;
+  /// let mut av = array_vec!([i32; 7] => 1, 2, 3);
+  /// let mut av2 = array_vec!([i32; 7] => 4, 5, 6);
+  /// av.append(&mut av2);
+  /// assert_eq!(av, &[1, 2, 3, 4, 5, 6][..]);
+  /// assert_eq!(av2, &[][..]);
+  ///
+  /// let mut av3 = array_vec!([i32; 7] => 7, 8, 9);
+  /// assert!(av.try_append(&mut av3).is_some());
+  /// assert_eq!(av, &[1, 2, 3, 4, 5, 6][..]);
+  /// assert_eq!(av3, &[7, 8, 9][..]);
+  /// ```
+  #[inline]
+  pub fn try_append<'other>(
+    &mut self,
+    other: &'other mut Self,
+  ) -> Option<&'other mut Self> {
+    let new_len = self.len() + other.len();
+    if new_len > A::CAPACITY {
+      return Some(other);
+    }
+
+    for item in other.drain(..) {
+      self.push(item);
+    }
+
+    return None;
+  }
+
   /// A `*mut` pointer to the backing array.
   ///
   /// ## Safety
@@ -197,7 +235,10 @@ impl<A: Array> ArrayVec<A> {
   #[inline(always)]
   #[must_use]
   pub fn capacity(&self) -> usize {
-    A::CAPACITY
+    // Note: This shouldn't use A::CAPACITY, because unsafe code can't rely on
+    // any Array invariants. This ensures that at the very least, the returned
+    // value is a valid length for a subslice of the backing array.
+    self.data.as_slice().len()
   }
 
   /// Truncates the `ArrayVec` down to length 0.
@@ -216,7 +257,7 @@ impl<A: Array> ArrayVec<A> {
   /// ## Example
   /// ```rust
   /// # use tinyvec::*;
-  /// let mut av = array_vec!([i32; 4], 1, 2, 3);
+  /// let mut av = array_vec!([i32; 4] => 1, 2, 3);
   /// let av2: ArrayVec<[i32; 4]> = av.drain(1..).collect();
   /// assert_eq!(av.as_slice(), &[1][..]);
   /// assert_eq!(av2.as_slice(), &[2, 3][..]);
@@ -274,35 +315,74 @@ impl<A: Array> ArrayVec<A> {
     }
 
     let new_len = self.len + sli.len();
-    if new_len > A::CAPACITY {
-      panic!(
-        "ArrayVec::extend_from_slice> total length {} exceeds capacity {}!",
-        new_len,
-        A::CAPACITY
-      )
-    }
+    assert!(
+      new_len <= A::CAPACITY,
+      "ArrayVec::extend_from_slice> total length {} exceeds capacity {}!",
+      new_len,
+      A::CAPACITY
+    );
 
     let target = &mut self.data.as_slice_mut()[self.len..new_len];
     target.clone_from_slice(sli);
     self.set_len(new_len);
   }
 
+  /// Fill the vector until its capacity has been reached.
+  ///
+  /// Successively fills unused space in the spare slice of the vector with
+  /// elements from the iterator. It then returns the remaining iterator
+  /// without exhausting it. This also allows appending the head of an
+  /// infinite iterator.
+  ///
+  /// This is an alternative to `Extend::extend` method for cases where the
+  /// length of the iterator can not be checked. Since this vector can not
+  /// reallocate to increase its capacity, it is unclear what to do with
+  /// remaining elements in the iterator and the iterator itself. The
+  /// interface also provides no way to communicate this to the caller.
+  ///
+  /// ## Panics
+  /// * If the `next` method of the provided iterator panics.
+  ///
+  /// ## Example
+  ///
+  /// ```rust
+  /// # use tinyvec::*;
+  /// let mut av = array_vec!([i32; 4]);
+  /// let mut to_inf = av.fill(0..);
+  /// assert_eq!(&av[..], [0, 1, 2, 3]);
+  /// assert_eq!(to_inf.next(), Some(4));
+  /// ```
+  #[inline]
+  pub fn fill<I: IntoIterator<Item = A::Item>>(
+    &mut self,
+    iter: I,
+  ) -> I::IntoIter {
+    let mut iter = iter.into_iter();
+    for element in iter.by_ref().take(self.capacity() - self.len()) {
+      self.push(element);
+    }
+    iter
+  }
+
   /// Wraps up an array and uses the given length as the initial length.
   ///
   /// If you want to simply use the full array, use `from` instead.
   ///
   /// ## Panics
   ///
-  /// * The length specified must be less than or equal to the capacity of the array.
+  /// * The length specified must be less than or equal to the capacity of the
+  ///   array.
   #[inline]
   #[must_use]
   #[allow(clippy::match_wild_err_arm)]
   pub fn from_array_len(data: A, len: usize) -> Self {
     match Self::try_from_array_len(data, len) {
       Ok(out) => out,
-      Err(_) => {
-        panic!("ArrayVec::from_array_len> length {} exceeds capacity {}!", len, A::CAPACITY)
-      }
+      Err(_) => panic!(
+        "ArrayVec::from_array_len> length {} exceeds capacity {}!",
+        len,
+        A::CAPACITY
+      ),
     }
   }
 
@@ -310,13 +390,13 @@ impl<A: Array> ArrayVec<A> {
   /// index.
   ///
   /// ## Panics
-  /// * If `index` > `len` or
+  /// * If `index` > `len`
   /// * If the capacity is exhausted
   ///
   /// ## Example
   /// ```rust
   /// use tinyvec::*;
-  /// let mut av = array_vec!([i32; 10], 1, 2, 3);
+  /// let mut av = array_vec!([i32; 10] => 1, 2, 3);
   /// av.insert(1, 4);
   /// assert_eq!(av.as_slice(), &[1, 4, 2, 3]);
   /// av.insert(4, 5);
@@ -324,14 +404,41 @@ impl<A: Array> ArrayVec<A> {
   /// ```
   #[inline]
   pub fn insert(&mut self, index: usize, item: A::Item) {
-    if index > self.len {
-      panic!("ArrayVec::insert> index {} is out of bounds {}", index, self.len);
+    let x = self.try_insert(index, item);
+    assert!(x.is_none(), "ArrayVec::insert> capacity overflow!");
+  }
+
+  /// Tries to insert an item at the position given, moving all following
+  /// elements +1 index.
+  /// Returns back the element if the capacity is exhausted,
+  /// otherwise returns None.
+  ///
+  /// ## Panics
+  /// * If `index` > `len`
+  ///
+  /// ## Example
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut av = array_vec!([&'static str; 4] => "one", "two", "three");
+  /// av.insert(1, "four");
+  /// assert_eq!(av.as_slice(), &["one", "four", "two", "three"]);
+  /// assert_eq!(av.try_insert(4, "five"), Some("five"));
+  /// ```
+  #[inline]
+  pub fn try_insert(&mut self, index: usize, item: A::Item) -> Option<A::Item> {
+    assert!(
+      index <= self.len,
+      "ArrayVec::try_insert> index {} is out of bounds {}",
+      index,
+      self.len
+    );
+
+    if let Some(x) = self.try_push(item) {
+      return Some(x);
     }
 
-    // Try to push the element.
-    self.push(item);
-    // And move it into its place.
     self.as_mut_slice()[index..].rotate_right(1);
+    return None;
   }
 
   /// Checks if the length is 0.
@@ -366,7 +473,7 @@ impl<A: Array> ArrayVec<A> {
   /// ## Example
   /// ```rust
   /// # use tinyvec::*;
-  /// let mut av = array_vec!([i32; 10], 1, 2);
+  /// let mut av = array_vec!([i32; 10] => 1, 2);
   /// assert_eq!(av.pop(), Some(2));
   /// assert_eq!(av.pop(), Some(1));
   /// assert_eq!(av.pop(), None);
@@ -400,12 +507,32 @@ impl<A: Array> ArrayVec<A> {
   /// ```
   #[inline(always)]
   pub fn push(&mut self, val: A::Item) {
-    if self.len < A::CAPACITY {
-      replace(&mut self.data.as_slice_mut()[self.len], val);
-      self.len += 1;
-    } else {
-      panic!("ArrayVec::push> capacity overflow!")
+    let x = self.try_push(val);
+    assert!(x.is_none(), "ArrayVec::push> capacity overflow!");
+  }
+
+  /// Tries to place an element onto the end of the vec.\
+  /// Returns back the element if the capacity is exhausted,
+  /// otherwise returns None.
+  /// ```rust
+  /// # use tinyvec::*;
+  /// let mut av = array_vec!([i32; 2]);
+  /// assert_eq!(av.as_slice(), []);
+  /// assert_eq!(av.try_push(1), None);
+  /// assert_eq!(&av[..], [1]);
+  /// assert_eq!(av.try_push(2), None);
+  /// assert_eq!(&av[..], [1, 2]);
+  /// assert_eq!(av.try_push(3), Some(3));
+  /// ```
+  #[inline(always)]
+  pub fn try_push(&mut self, val: A::Item) -> Option<A::Item> {
+    if self.len == A::CAPACITY {
+      return Some(val);
     }
+
+    self.data.as_slice_mut()[self.len] = val;
+    self.len += 1;
+    return None;
   }
 
   /// Removes the item at `index`, shifting all others down by one index.
@@ -420,34 +547,32 @@ impl<A: Array> ArrayVec<A> {
   ///
   /// ```rust
   /// # use tinyvec::*;
-  /// let mut av = array_vec!([i32; 4], 1, 2, 3);
+  /// let mut av = array_vec!([i32; 4] => 1, 2, 3);
   /// assert_eq!(av.remove(1), 2);
   /// assert_eq!(&av[..], [1, 3]);
   /// ```
   #[inline]
   pub fn remove(&mut self, index: usize) -> A::Item {
     let targets: &mut [A::Item] = &mut self.deref_mut()[index..];
-    let item = replace(&mut targets[0], A::Item::default());
+    let item = take(&mut targets[0]);
     targets.rotate_left(1);
     self.len -= 1;
     item
   }
 
-  /// Resize the vec to the new length.
-  ///
-  /// If it needs to be longer, it's filled with clones of the provided value.
-  /// If it needs to be shorter, it's truncated.
+  /// As [`resize_with`](ArrayVec::resize_with)
+  /// and it clones the value as the closure.
   ///
   /// ## Example
   ///
   /// ```rust
   /// # use tinyvec::*;
   ///
-  /// let mut av = array_vec!([&str; 10], "hello");
+  /// let mut av = array_vec!([&str; 10] => "hello");
   /// av.resize(3, "world");
   /// assert_eq!(&av[..], ["hello", "world", "world"]);
   ///
-  /// let mut av = array_vec!([i32; 10], 1, 2, 3, 4);
+  /// let mut av = array_vec!([i32; 10] => 1, 2, 3, 4);
   /// av.resize(2, 0);
   /// assert_eq!(&av[..], [1, 2]);
   /// ```
@@ -456,16 +581,7 @@ impl<A: Array> ArrayVec<A> {
   where
     A::Item: Clone,
   {
-    match new_len.checked_sub(self.len) {
-      None => self.truncate(new_len),
-      Some(0) => (),
-      Some(new_elements) => {
-        for _ in 1..new_elements {
-          self.push(new_val.clone());
-        }
-        self.push(new_val);
-      }
-    }
+    self.resize_with(new_len, || new_val.clone())
   }
 
   /// Resize the vec to the new length.
@@ -478,13 +594,16 @@ impl<A: Array> ArrayVec<A> {
   /// ```rust
   /// # use tinyvec::*;
   ///
-  /// let mut av = array_vec!([i32; 10], 1, 2, 3);
+  /// let mut av = array_vec!([i32; 10] => 1, 2, 3);
   /// av.resize_with(5, Default::default);
   /// assert_eq!(&av[..], [1, 2, 3, 0, 0]);
   ///
   /// let mut av = array_vec!([i32; 10]);
   /// let mut p = 1;
-  /// av.resize_with(4, || { p *= 2; p });
+  /// av.resize_with(4, || {
+  ///   p *= 2;
+  ///   p
+  /// });
   /// assert_eq!(&av[..], [2, 4, 8, 16]);
   /// ```
   #[inline]
@@ -510,7 +629,7 @@ impl<A: Array> ArrayVec<A> {
   /// ```rust
   /// # use tinyvec::*;
   ///
-  /// let mut av = array_vec!([i32; 10], 1, 1, 2, 3, 3, 4);
+  /// let mut av = array_vec!([i32; 10] => 1, 1, 2, 3, 3, 4);
   /// av.retain(|&x| x % 2 == 0);
   /// assert_eq!(&av[..], [2, 4]);
   /// ```
@@ -567,49 +686,16 @@ impl<A: Array> ArrayVec<A> {
       // just let some other call later on trigger a panic on accident when the
       // length is wrong. However, it's a lot easier to catch bugs when things
       // are more "fail-fast".
-      panic!("ArrayVec: set_len overflow!")
+      panic!(
+        "ArrayVec::set_len> new length {} exceeds capacity {}",
+        new_len,
+        A::CAPACITY
+      )
     } else {
       self.len = new_len;
     }
   }
 
-  /// Fill the vector until its capacity has been reached.
-  ///
-  /// Successively fills unused space in the spare slice of the vector with
-  /// elements from the iterator. It then returns the remaining iterator
-  /// without exhausting it. This also allows appending the head of an
-  /// infinite iterator.
-  ///
-  /// This is an alternative to `Extend::extend` method for cases where the
-  /// length of the iterator can not be checked. Since this vector can not
-  /// reallocate to increase its capacity, it is unclear what to do with
-  /// remaining elements in the iterator and the iterator itself. The
-  /// interface also provides no way to communicate this to the caller.
-  ///
-  /// ## Panics
-  /// * If the `next` method of the provided iterator panics.
-  ///
-  /// ## Example
-  ///
-  /// ```rust
-  /// # use tinyvec::*;
-  /// let mut av = array_vec!([i32; 4]);
-  /// let mut to_inf = av.fill(0..);
-  /// assert_eq!(&av[..], [0, 1, 2, 3]);
-  /// assert_eq!(to_inf.next(), Some(4));
-  /// ```
-  #[inline]
-  pub fn fill<I: IntoIterator<Item = A::Item>>(
-    &mut self,
-    iter: I,
-  ) -> I::IntoIter {
-    let mut iter = iter.into_iter();
-    for element in iter.by_ref().take(self.capacity() - self.len()) {
-      self.push(element);
-    }
-    iter
-  }
-
   /// Splits the collection at the point given.
   ///
   /// * `[0, at)` stays in this vec
@@ -622,7 +708,7 @@ impl<A: Array> ArrayVec<A> {
   ///
   /// ```rust
   /// # use tinyvec::*;
-  /// let mut av = array_vec!([i32; 4], 1, 2, 3);
+  /// let mut av = array_vec!([i32; 4] => 1, 2, 3);
   /// let av2 = av.split_off(1);
   /// assert_eq!(&av[..], [1]);
   /// assert_eq!(&av2[..], [2, 3]);
@@ -649,6 +735,74 @@ impl<A: Array> ArrayVec<A> {
     new
   }
 
+  /// Creates a splicing iterator that removes the specified range in the
+  /// vector, yields the removed items, and replaces them with elements from
+  /// the provided iterator.
+  ///
+  /// `splice` fuses the provided iterator, so elements after the first `None`
+  /// are ignored.
+  ///
+  /// ## Panics
+  /// * If the start is greater than the end.
+  /// * If the end is past the edge of the vec.
+  /// * If the provided iterator panics.
+  /// * If the new length would overflow the capacity of the array. Because
+  ///   `ArrayVecSplice` adds elements to this vec in its destructor when
+  ///   necessary, this panic would occur when it is dropped.
+  ///
+  /// ## Example
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut av = array_vec!([i32; 4] => 1, 2, 3);
+  /// let av2: ArrayVec<[i32; 4]> = av.splice(1.., 4..=6).collect();
+  /// assert_eq!(av.as_slice(), &[1, 4, 5, 6][..]);
+  /// assert_eq!(av2.as_slice(), &[2, 3][..]);
+  ///
+  /// av.splice(.., None);
+  /// assert_eq!(av.as_slice(), &[]);
+  /// ```
+  #[inline]
+  pub fn splice<R, I>(
+    &mut self,
+    range: R,
+    replacement: I,
+  ) -> ArrayVecSplice<'_, A, core::iter::Fuse<I::IntoIter>>
+  where
+    R: RangeBounds<usize>,
+    I: IntoIterator<Item = A::Item>,
+  {
+    use core::ops::Bound;
+    let start = match range.start_bound() {
+      Bound::Included(x) => *x,
+      Bound::Excluded(x) => x + 1,
+      Bound::Unbounded => 0,
+    };
+    let end = match range.end_bound() {
+      Bound::Included(x) => x + 1,
+      Bound::Excluded(x) => *x,
+      Bound::Unbounded => self.len(),
+    };
+    assert!(
+      start <= end,
+      "ArrayVec::splice> Illegal range, {} to {}",
+      start,
+      end
+    );
+    assert!(
+      end <= self.len(),
+      "ArrayVec::splice> Range ends at {} but length is only {}!",
+      end,
+      self.len()
+    );
+
+    ArrayVecSplice {
+      removal_start: start,
+      removal_end: end,
+      parent: self,
+      replacement: replacement.into_iter().fuse(),
+    }
+  }
+
   /// Remove an element, swapping the end of the vec into its place.
   ///
   /// ## Panics
@@ -657,7 +811,7 @@ impl<A: Array> ArrayVec<A> {
   /// ## Example
   /// ```rust
   /// # use tinyvec::*;
-  /// let mut av = array_vec!([&str; 4], "foo", "bar", "quack", "zap");
+  /// let mut av = array_vec!([&str; 4] => "foo", "bar", "quack", "zap");
   ///
   /// assert_eq!(av.swap_remove(1), "bar");
   /// assert_eq!(&av[..], ["foo", "zap", "quack"]);
@@ -786,7 +940,7 @@ impl<A: Array> ArrayVec<A> {
   }
 }
 
-/// Draining iterator for `ArrayVecDrain`
+/// Draining iterator for [`ArrayVec`]
 ///
 /// See [`ArrayVec::drain`](ArrayVec::drain)
 pub struct ArrayVecDrain<'p, A: Array> {
@@ -808,19 +962,126 @@ impl<'p, A: Array> Iterator for ArrayVecDrain<'p, A> {
     }
   }
 }
+impl<'p, A: Array> FusedIterator for ArrayVecDrain<'p, A> {}
 impl<'p, A: Array> Drop for ArrayVecDrain<'p, A> {
   #[inline]
   fn drop(&mut self) {
-    // Changed because it was moving `self`, it's also more clear and the std does the same
+    // Changed because it was moving `self`, it's also more clear and the std
+    // does the same
     self.for_each(drop);
     // Implementation very similar to [`ArrayVec::remove`](ArrayVec::remove)
     let count = self.target_end - self.target_start;
-    let targets: &mut [A::Item] = &mut self.parent.deref_mut()[self.target_start..];
+    let targets: &mut [A::Item] =
+      &mut self.parent.deref_mut()[self.target_start..];
     targets.rotate_left(count);
     self.parent.len -= count;
   }
 }
 
+/// Splicing iterator for `ArrayVec`
+/// See [`ArrayVec::splice`](ArrayVec::<A>::splice)
+pub struct ArrayVecSplice<'p, A: Array, I: Iterator<Item = A::Item>> {
+  parent: &'p mut ArrayVec<A>,
+  removal_start: usize,
+  removal_end: usize,
+  replacement: I,
+}
+
+impl<'p, A: Array, I: Iterator<Item = A::Item>> Iterator
+  for ArrayVecSplice<'p, A, I>
+{
+  type Item = A::Item;
+
+  fn next(&mut self) -> Option<A::Item> {
+    if self.removal_start < self.removal_end {
+      match self.replacement.next() {
+        Some(replacement) => {
+          let removed = core::mem::replace(
+            &mut self.parent[self.removal_start],
+            replacement,
+          );
+          self.removal_start += 1;
+          Some(removed)
+        }
+        None => {
+          let removed = self.parent.remove(self.removal_start);
+          self.removal_end -= 1;
+          Some(removed)
+        }
+      }
+    } else {
+      None
+    }
+  }
+
+  #[inline]
+  fn size_hint(&self) -> (usize, Option<usize>) {
+    let len = self.len();
+    (len, Some(len))
+  }
+}
+
+impl<'p, A, I> ExactSizeIterator for ArrayVecSplice<'p, A, I>
+where
+  A: Array,
+  I: Iterator<Item = A::Item>,
+{
+  #[inline]
+  fn len(&self) -> usize {
+    self.removal_end - self.removal_start
+  }
+}
+
+impl<'p, A, I> FusedIterator for ArrayVecSplice<'p, A, I>
+where
+  A: Array,
+  I: Iterator<Item = A::Item>,
+{
+}
+
+impl<'p, A, I> DoubleEndedIterator for ArrayVecSplice<'p, A, I>
+where
+  A: Array,
+  I: Iterator<Item = A::Item> + DoubleEndedIterator,
+{
+  fn next_back(&mut self) -> Option<A::Item> {
+    if self.removal_start < self.removal_end {
+      match self.replacement.next_back() {
+        Some(replacement) => {
+          let removed = core::mem::replace(
+            &mut self.parent[self.removal_end - 1],
+            replacement,
+          );
+          self.removal_end -= 1;
+          Some(removed)
+        }
+        None => {
+          let removed = self.parent.remove(self.removal_end - 1);
+          self.removal_end -= 1;
+          Some(removed)
+        }
+      }
+    } else {
+      None
+    }
+  }
+}
+
+impl<'p, A: Array, I: Iterator<Item = A::Item>> Drop
+  for ArrayVecSplice<'p, A, I>
+{
+  fn drop(&mut self) {
+    for _ in self.by_ref() {}
+
+    // FIXME: reserve lower bound of size_hint
+
+    for replacement in self.replacement.by_ref() {
+      self.parent.insert(self.removal_end, replacement);
+      self.removal_end += 1;
+    }
+  }
+}
+
 impl<A: Array> AsMut<[A::Item]> for ArrayVec<A> {
   #[inline(always)]
   #[must_use]
@@ -901,7 +1162,7 @@ impl<A: Array> ArrayVecIterator<A> {
     &self.data.as_slice()[self.base..self.len]
   }
 }
-
+impl<A: Array> FusedIterator for ArrayVecIterator<A> {}
 impl<A: Array> Iterator for ArrayVecIterator<A> {
   type Item = A::Item;
   #[inline]
@@ -941,7 +1202,10 @@ impl<A: Array> Iterator for ArrayVecIterator<A> {
   }
 }
 
-impl<A: Array> Debug for ArrayVecIterator<A> where A::Item: Debug {
+impl<A: Array> Debug for ArrayVecIterator<A>
+where
+  A::Item: Debug,
+{
   #[allow(clippy::missing_inline_in_public_items)]
   fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
     f.debug_tuple("ArrayVecIterator").field(&self.as_slice()).finish()
@@ -958,6 +1222,26 @@ impl<A: Array> IntoIterator for ArrayVec<A> {
   }
 }
 
+impl<'a, A: Array> IntoIterator for &'a mut ArrayVec<A> {
+  type Item = &'a mut A::Item;
+  type IntoIter = core::slice::IterMut<'a, A::Item>;
+  #[inline(always)]
+  #[must_use]
+  fn into_iter(self) -> Self::IntoIter {
+    self.iter_mut()
+  }
+}
+
+impl<'a, A: Array> IntoIterator for &'a ArrayVec<A> {
+  type Item = &'a A::Item;
+  type IntoIter = core::slice::Iter<'a, A::Item>;
+  #[inline(always)]
+  #[must_use]
+  fn into_iter(self) -> Self::IntoIter {
+    self.iter()
+  }
+}
+
 impl<A: Array> PartialEq for ArrayVec<A>
 where
   A::Item: PartialEq,
@@ -1023,6 +1307,22 @@ where
   }
 }
 
+#[cfg(feature = "experimental_write_impl")]
+impl<A: Array<Item = u8>> core::fmt::Write for ArrayVec<A> {
+  fn write_str(&mut self, s: &str) -> core::fmt::Result {
+    let my_len = self.len();
+    let str_len = s.as_bytes().len();
+    if my_len + str_len <= A::CAPACITY {
+      let remainder = &mut self.data.as_slice_mut()[my_len..];
+      let target = &mut remainder[..str_len];
+      target.copy_from_slice(s.as_bytes());
+      Ok(())
+    } else {
+      Err(core::fmt::Error)
+    }
+  }
+}
+
 // // // // // // // //
 // Formatting impls
 // // // // // // // //
@@ -1179,3 +1479,36 @@ where
     write!(f, "]")
   }
 }
+
+#[cfg(feature = "alloc")]
+use alloc::vec::Vec;
+
+#[cfg(feature = "alloc")]
+impl<A: Array> ArrayVec<A> {
+  /// Drains all elements to a Vec, but reserves additional space
+  /// ```
+  /// # use tinyvec::*;
+  /// let mut av = array_vec!([i32; 7] => 1, 2, 3);
+  /// let v = av.drain_to_vec_and_reserve(10);
+  /// assert_eq!(v, &[1, 2, 3]);
+  /// assert_eq!(v.capacity(), 13);
+  /// ```
+  pub fn drain_to_vec_and_reserve(&mut self, n: usize) -> Vec<A::Item> {
+    let cap = n + self.len();
+    let mut v = Vec::with_capacity(cap);
+    v.extend(self.drain(..));
+    return v;
+  }
+
+  /// Drains all elements to a Vec
+  /// ```
+  /// # use tinyvec::*;
+  /// let mut av = array_vec!([i32; 7] => 1, 2, 3);
+  /// let v = av.drain_to_vec();
+  /// assert_eq!(v, &[1, 2, 3]);
+  /// assert_eq!(v.capacity(), 3);
+  /// ```
+  pub fn drain_to_vec(&mut self) -> Vec<A::Item> {
+    self.drain_to_vec_and_reserve(0)
+  }
+}
