@@ -17,8 +17,8 @@ use alloc::vec::Vec;
 ///
 /// // The backing array type can be specified in the macro call
 /// let empty_tv = tiny_vec!([u8; 16]);
-/// let some_ints = tiny_vec!([i32; 4], 1, 2, 3);
-/// let many_ints = tiny_vec!([i32; 4], 1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
+/// let some_ints = tiny_vec!([i32; 4] => 1, 2, 3);
+/// let many_ints = tiny_vec!([i32; 4] => 1, 2, 3, 4, 5, 6, 7, 8, 9, 10);
 ///
 /// // Or left to inference
 /// let empty_tv: TinyVec<[u8; 16]> = tiny_vec!();
@@ -27,38 +27,37 @@ use alloc::vec::Vec;
 /// ```
 #[macro_export]
 macro_rules! tiny_vec {
-  ($array_type:ty) => {
+  ($array_type:ty => $($elem:expr),* $(,)?) => {
     {
-      let mut tv: $crate::TinyVec<$array_type> = Default::default();
-      tv
-    }
-  };
-  ($array_type:ty, $($elem:expr),* $(,)?) => {
-    {
-      // Note(Lokathor): This goofy looking thing will count the number of
-      // `$elem` entries we were given. We can't spit out the "+1"s on their
-      // own, we need to use `$elem` in the repetition-expansion somehow.
-      // However, we also can't assume it's `Copy` data, so we must use `$elem`
-      // only once "for real" in the expansion as a whole. To achieve this, we
-      // can `stringify!` each element in an inner block, then have the block
-      // return a 1. The stringification is a compile time thing, it won't
-      // actually move any values.
+      // https://github.com/rust-lang/lang-team/issues/28
       const INVOKED_ELEM_COUNT: usize = 0 $( + { let _ = stringify!($elem); 1 })*;
       // If we have more `$elem` than the `CAPACITY` we will simply go directly
       // to constructing on the heap.
-      let av: $crate::TinyVec<$array_type> = $crate::TinyVec::from_either_with_capacity(
-        INVOKED_ELEM_COUNT,
-        #[inline(always)] || $crate::array_vec!($array_type, $($elem),*),
-        #[inline(always)] || vec!($($elem),*));
-      av
+      match $crate::TinyVec::constructor_for_capacity(INVOKED_ELEM_COUNT) {
+        $crate::TinyVecConstructor::Inline(f) => {
+          f($crate::array_vec!($array_type => $($elem),*))
+        }
+        $crate::TinyVecConstructor::Heap(f) => {
+          f(vec!($($elem),*))
+        }
+      }
     }
   };
-  () => {
-    tiny_vec!(_)
+  ($array_type:ty) => {
+    $crate::tiny_vec!($array_type =>)
   };
   ($($elem:expr),*) => {
-    tiny_vec!(_, $($elem),*)
+    $crate::tiny_vec!(_ => $($elem),*)
   };
+  () => {
+    $crate::tiny_vec!(_)
+  };
+}
+
+#[doc(hidden)] // Internal implementation details of `tiny_vec!`
+pub enum TinyVecConstructor<A: Array> {
+  Inline(fn(ArrayVec<A>) -> TinyVec<A>),
+  Heap(fn(Vec<A::Item>) -> TinyVec<A>),
 }
 
 /// A vector that starts inline, but can automatically move to the heap.
@@ -80,7 +79,7 @@ macro_rules! tiny_vec {
 /// ```rust
 /// # use tinyvec::*;
 /// let empty_tv = tiny_vec!([u8; 16]);
-/// let some_ints = tiny_vec!([i32; 4], 1, 2, 3);
+/// let some_ints = tiny_vec!([i32; 4] => 1, 2, 3);
 /// ```
 #[derive(Clone)]
 pub enum TinyVec<A: Array> {
@@ -138,19 +137,144 @@ impl<A: Array, I: SliceIndex<[A::Item]>> IndexMut<I> for TinyVec<A> {
 }
 
 impl<A: Array> TinyVec<A> {
+  /// Returns whether elements are on heap
+  pub fn is_heap(&self) -> bool {
+    match self {
+      TinyVec::Heap(_) => true,
+      TinyVec::Inline(_) => false,
+    }
+  }
+  /// Returns whether elements are on stack
+  pub fn is_inline(&self) -> bool {
+    !self.is_heap()
+  }
+
+  /// Shrinks the capacity of the vector as much as possible.\
+  /// It is inlined if length is less than `A::CAPACITY`.
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 2] => 1, 2, 3);
+  /// assert!(tv.is_heap());
+  /// let _ = tv.pop();
+  /// assert!(tv.is_heap());
+  /// tv.shrink_to_fit();
+  /// assert!(tv.is_inline());
+  /// ```
+  pub fn shrink_to_fit(&mut self)
+  where
+    A: Default,
+  {
+    let vec = match self {
+      TinyVec::Inline(_) => return,
+      TinyVec::Heap(h) => h,
+    };
+
+    if vec.len() > A::CAPACITY {
+      return vec.shrink_to_fit();
+    }
+
+    let movedvec = core::mem::replace(vec, Vec::new());
+
+    let mut av = ArrayVec::default();
+    let mut rest = av.fill(movedvec);
+    debug_assert!(rest.next().is_none());
+    *self = TinyVec::Inline(av);
+  }
+
   /// Moves the content of the TinyVec to the heap, if it's inline.
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
+  /// assert!(tv.is_inline());
+  /// tv.move_to_the_heap();
+  /// assert!(tv.is_heap());
+  /// ```
   #[allow(clippy::missing_inline_in_public_items)]
   pub fn move_to_the_heap(&mut self) {
-    match self {
-      TinyVec::Inline(ref mut arr) => {
-        let mut v = Vec::with_capacity(A::CAPACITY * 2);
-        for item in arr.drain(..) {
-          v.push(item);
-        }
-        replace(self, TinyVec::Heap(v));
-      }
-      TinyVec::Heap(_) => (),
+    let arr = match self {
+      TinyVec::Heap(_) => return,
+      TinyVec::Inline(a) => a,
+    };
+
+    let v = arr.drain_to_vec();
+    *self = TinyVec::Heap(v);
+  }
+
+  /// If TinyVec is inline, moves the content of it to the heap.
+  /// Also reserves additional space.
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
+  /// assert!(tv.is_inline());
+  /// tv.move_to_the_heap_and_reserve(32);
+  /// assert!(tv.is_heap());
+  /// assert!(tv.capacity() >= 35);
+  /// ```
+  pub fn move_to_the_heap_and_reserve(&mut self, n: usize) {
+    let arr = match self {
+      TinyVec::Heap(h) => return h.reserve(n),
+      TinyVec::Inline(a) => a,
+    };
+
+    let v = arr.drain_to_vec_and_reserve(n);
+    *self = TinyVec::Heap(v);
+  }
+
+  /// Reserves additional space.
+  /// Moves to the heap if array can't hold `n` more items
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3, 4);
+  /// assert!(tv.is_inline());
+  /// tv.reserve(1);
+  /// assert!(tv.is_heap());
+  /// assert!(tv.capacity() >= 5);
+  /// ```
+  pub fn reserve(&mut self, n: usize) {
+    let arr = match self {
+      TinyVec::Heap(h) => return h.reserve(n),
+      TinyVec::Inline(a) => a,
+    };
+
+    if n > arr.capacity() - arr.len() {
+      let v = arr.drain_to_vec_and_reserve(n);
+      *self = TinyVec::Heap(v);
     }
+
+    /* In this place array has enough place, so no work is needed more */
+    return;
+  }
+
+  /// Reserves additional space.
+  /// Moves to the heap if array can't hold `n` more items
+  ///
+  /// From [Vec::reserve_exact](https://doc.rust-lang.org/std/vec/struct.Vec.html#method.reserve_exact)
+  /// ```text
+  /// Note that the allocator may give the collection more space than it requests.
+  /// Therefore, capacity can not be relied upon to be precisely minimal.
+  /// Prefer reserve if future insertions are expected.
+  /// ```
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3, 4);
+  /// assert!(tv.is_inline());
+  /// tv.reserve_exact(1);
+  /// assert!(tv.is_heap());
+  /// assert!(tv.capacity() >= 5);
+  /// ```
+  pub fn reserve_exact(&mut self, n: usize) {
+    let arr = match self {
+      TinyVec::Heap(h) => return h.reserve_exact(n),
+      TinyVec::Inline(a) => a,
+    };
+
+    if n > arr.capacity() - arr.len() {
+      let v = arr.drain_to_vec_and_reserve(n);
+      *self = TinyVec::Heap(v);
+    }
+
+    /* In this place array has enough place, so no work is needed more */
+    return;
   }
 }
 
@@ -158,8 +282,12 @@ impl<A: Array> TinyVec<A> {
   /// Move all values from `other` into this vec.
   #[inline]
   pub fn append(&mut self, other: &mut Self) {
-    for item in other.drain(..) {
-      self.push(item)
+    self.reserve(other.len());
+    let iter = other.drain(..);
+
+    match self {
+      TinyVec::Heap(h) => h.extend(iter),
+      TinyVec::Inline(a) => a.extend(iter),
     }
   }
 
@@ -213,7 +341,8 @@ impl<A: Array> TinyVec<A> {
   #[must_use]
   pub fn capacity(&self) -> usize {
     match self {
-      TinyVec::Inline(_) => A::CAPACITY,
+      // Note: this shouldn't use A::CAPACITY. See ArrayVec::capacity().
+      TinyVec::Inline(v) => v.capacity(),
       TinyVec::Heap(v) => v.capacity(),
     }
   }
@@ -269,7 +398,7 @@ impl<A: Array> TinyVec<A> {
   /// ## Example
   /// ```rust
   /// use tinyvec::*;
-  /// let mut tv = tiny_vec!([i32; 4], 1, 2, 3);
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
   /// let tv2: TinyVec<[i32; 4]> = tv.drain(1..).collect();
   /// assert_eq!(tv.as_slice(), &[1][..]);
   /// assert_eq!(tv2.as_slice(), &[2, 3][..]);
@@ -313,13 +442,21 @@ impl<A: Array> TinyVec<A> {
   }
 
   /// Clone each element of the slice into this vec.
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2);
+  /// tv.extend_from_slice(&[3, 4]);
+  /// assert_eq!(tv.as_slice(), [1, 2, 3, 4]);
+  /// ```
   #[inline]
   pub fn extend_from_slice(&mut self, sli: &[A::Item])
   where
     A::Item: Clone,
   {
-    for i in sli {
-      self.push(i.clone())
+    self.reserve(sli.len());
+    match self {
+      TinyVec::Inline(a) => a.extend_from_slice(sli),
+      TinyVec::Heap(h) => h.extend_from_slice(sli),
     }
   }
 
@@ -344,15 +481,11 @@ impl<A: Array> TinyVec<A> {
 
   #[inline(always)]
   #[doc(hidden)] // Internal implementation details of `tiny_vec!`
-  pub fn from_either_with_capacity(
-    cap: usize,
-    make_array: impl FnOnce() -> ArrayVec<A>,
-    make_vec: impl FnOnce() -> Vec<A::Item>,
-  ) -> Self {
+  pub fn constructor_for_capacity(cap: usize) -> TinyVecConstructor<A> {
     if cap <= A::CAPACITY {
-      TinyVec::Inline(make_array())
+      TinyVecConstructor::Inline(TinyVec::Inline)
     } else {
-      TinyVec::Heap(make_vec())
+      TinyVecConstructor::Heap(TinyVec::Heap)
     }
   }
 
@@ -365,7 +498,7 @@ impl<A: Array> TinyVec<A> {
   /// ## Example
   /// ```rust
   /// use tinyvec::*;
-  /// let mut tv = tiny_vec!([i32; 10], 1, 2, 3);
+  /// let mut tv = tiny_vec!([i32; 10] => 1, 2, 3);
   /// tv.insert(1, 4);
   /// assert_eq!(tv.as_slice(), &[1, 4, 2, 3]);
   /// tv.insert(4, 5);
@@ -373,16 +506,26 @@ impl<A: Array> TinyVec<A> {
   /// ```
   #[inline]
   pub fn insert(&mut self, index: usize, item: A::Item) {
-    match self {
-      TinyVec::Inline(a) => {
-        if a.len() == A::CAPACITY {
-          self.move_to_the_heap();
-          self.insert(index, item)
-        } else {
-          a.insert(index, item);
-        }
-      }
-      TinyVec::Heap(v) => v.insert(index, item),
+    assert!(
+      index <= self.len(),
+      "insertion index (is {}) should be <= len (is {})",
+      index,
+      self.len()
+    );
+
+    let arr = match self {
+      TinyVec::Heap(v) => return v.insert(index, item),
+      TinyVec::Inline(a) => a,
+    };
+
+    if let Some(x) = arr.try_insert(index, item) {
+      let mut v = Vec::with_capacity(arr.len() * 2);
+      let mut it =
+        arr.iter_mut().map(|r| core::mem::replace(r, Default::default()));
+      v.extend(it.by_ref().take(index));
+      v.push(x);
+      v.extend(it);
+      *self = TinyVec::Heap(v);
     }
   }
 
@@ -428,18 +571,24 @@ impl<A: Array> TinyVec<A> {
   /// Place an element onto the end of the vec.
   /// ## Panics
   /// * If the length of the vec would overflow the capacity.
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 10] => 1, 2, 3);
+  /// tv.push(4);
+  /// assert_eq!(tv.as_slice(), &[1, 2, 3, 4]);
+  /// ```
   #[inline(always)]
   pub fn push(&mut self, val: A::Item) {
-    match self {
-      TinyVec::Inline(a) => {
-        if a.len() == A::CAPACITY {
-          self.move_to_the_heap();
-          self.push(val)
-        } else {
-          a.push(val);
-        }
-      }
-      TinyVec::Heap(v) => v.push(val),
+    let arr = match self {
+      TinyVec::Heap(v) => return v.push(val),
+      TinyVec::Inline(a) => a,
+    };
+
+    if let Some(x) = arr.try_push(val) {
+      /* Make the Vec twice the size to amortize the cost of draining */
+      let mut v = arr.drain_to_vec_and_reserve(arr.len());
+      v.push(x);
+      *self = TinyVec::Heap(v);
     }
   }
 
@@ -455,7 +604,7 @@ impl<A: Array> TinyVec<A> {
   ///
   /// ```rust
   /// use tinyvec::*;
-  /// let mut tv = tiny_vec!([i32; 4], 1, 2, 3);
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
   /// assert_eq!(tv.remove(1), 2);
   /// assert_eq!(tv.as_slice(), &[1, 3][..]);
   /// ```
@@ -477,11 +626,11 @@ impl<A: Array> TinyVec<A> {
   /// ```rust
   /// use tinyvec::*;
   ///
-  /// let mut tv = tiny_vec!([&str; 10], "hello");
+  /// let mut tv = tiny_vec!([&str; 10] => "hello");
   /// tv.resize(3, "world");
   /// assert_eq!(tv.as_slice(), &["hello", "world", "world"][..]);
   ///
-  /// let mut tv = tiny_vec!([i32; 10], 1, 2, 3, 4);
+  /// let mut tv = tiny_vec!([i32; 10] => 1, 2, 3, 4);
   /// tv.resize(2, 0);
   /// assert_eq!(tv.as_slice(), &[1, 2][..]);
   /// ```
@@ -490,17 +639,7 @@ impl<A: Array> TinyVec<A> {
   where
     A::Item: Clone,
   {
-    match self {
-      TinyVec::Inline(a) => {
-        if new_len > A::CAPACITY {
-          self.move_to_the_heap();
-          self.resize(new_len, new_val);
-        } else {
-          a.resize(new_len, new_val);
-        }
-      }
-      TinyVec::Heap(v) => v.resize(new_len, new_val),
-    }
+    self.resize_with(new_len, || new_val.clone());
   }
 
   /// Resize the vec to the new length.
@@ -513,11 +652,11 @@ impl<A: Array> TinyVec<A> {
   /// ```rust
   /// use tinyvec::*;
   ///
-  /// let mut tv = tiny_vec!([i32; 10], 1, 2, 3);
+  /// let mut tv = tiny_vec!([i32; 3] => 1, 2, 3);
   /// tv.resize_with(5, Default::default);
   /// assert_eq!(tv.as_slice(), &[1, 2, 3, 0, 0][..]);
   ///
-  /// let mut tv = tiny_vec!([i32; 10]);
+  /// let mut tv = tiny_vec!([i32; 2]);
   /// let mut p = 1;
   /// tv.resize_with(4, || {
   ///   p *= 2;
@@ -527,6 +666,11 @@ impl<A: Array> TinyVec<A> {
   /// ```
   #[inline]
   pub fn resize_with<F: FnMut() -> A::Item>(&mut self, new_len: usize, f: F) {
+    match new_len.checked_sub(self.len()) {
+      None => return self.truncate(new_len),
+      Some(n) => self.reserve(n),
+    }
+
     match self {
       TinyVec::Inline(a) => a.resize_with(new_len, f),
       TinyVec::Heap(v) => v.resize_with(new_len, f),
@@ -540,7 +684,7 @@ impl<A: Array> TinyVec<A> {
   /// ```rust
   /// use tinyvec::*;
   ///
-  /// let mut tv = tiny_vec!([i32; 10], 1, 2, 3, 4);
+  /// let mut tv = tiny_vec!([i32; 10] => 1, 2, 3, 4);
   /// tv.retain(|&x| x % 2 == 0);
   /// assert_eq!(tv.as_slice(), &[2, 4][..]);
   /// ```
@@ -564,7 +708,7 @@ impl<A: Array> TinyVec<A> {
   ///
   /// ```rust
   /// use tinyvec::*;
-  /// let mut tv = tiny_vec!([i32; 4], 1, 2, 3);
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
   /// let tv2 = tv.split_off(1);
   /// assert_eq!(tv.as_slice(), &[1][..]);
   /// assert_eq!(tv2.as_slice(), &[2, 3][..]);
@@ -580,6 +724,71 @@ impl<A: Array> TinyVec<A> {
     }
   }
 
+  /// Creates a splicing iterator that removes the specified range in the
+  /// vector, yields the removed items, and replaces them with elements from
+  /// the provided iterator.
+  ///
+  /// `splice` fuses the provided iterator, so elements after the first `None`
+  /// are ignored.
+  ///
+  /// ## Panics
+  /// * If the start is greater than the end.
+  /// * If the end is past the edge of the vec.
+  /// * If the provided iterator panics.
+  ///
+  /// ## Example
+  /// ```rust
+  /// use tinyvec::*;
+  /// let mut tv = tiny_vec!([i32; 4] => 1, 2, 3);
+  /// let tv2: TinyVec<[i32; 4]> = tv.splice(1.., 4..=6).collect();
+  /// assert_eq!(tv.as_slice(), &[1, 4, 5, 6][..]);
+  /// assert_eq!(tv2.as_slice(), &[2, 3][..]);
+  ///
+  /// tv.splice(.., None);
+  /// assert_eq!(tv.as_slice(), &[]);
+  /// ```
+  #[inline]
+  pub fn splice<R, I>(
+    &mut self,
+    range: R,
+    replacement: I,
+  ) -> TinyVecSplice<'_, A, core::iter::Fuse<I::IntoIter>>
+  where
+    R: RangeBounds<usize>,
+    I: IntoIterator<Item = A::Item>,
+  {
+    use core::ops::Bound;
+    let start = match range.start_bound() {
+      Bound::Included(x) => *x,
+      Bound::Excluded(x) => x + 1,
+      Bound::Unbounded => 0,
+    };
+    let end = match range.end_bound() {
+      Bound::Included(x) => x + 1,
+      Bound::Excluded(x) => *x,
+      Bound::Unbounded => self.len(),
+    };
+    assert!(
+      start <= end,
+      "TinyVec::splice> Illegal range, {} to {}",
+      start,
+      end
+    );
+    assert!(
+      end <= self.len(),
+      "TinyVec::splice> Range ends at {} but length is only {}!",
+      end,
+      self.len()
+    );
+
+    TinyVecSplice {
+      removal_start: start,
+      removal_end: end,
+      parent: self,
+      replacement: replacement.into_iter().fuse(),
+    }
+  }
+
   /// Remove an element, swapping the end of the vec into its place.
   ///
   /// ## Panics
@@ -588,7 +797,7 @@ impl<A: Array> TinyVec<A> {
   /// ## Example
   /// ```rust
   /// use tinyvec::*;
-  /// let mut tv = tiny_vec!([&str; 4], "foo", "bar", "quack", "zap");
+  /// let mut tv = tiny_vec!([&str; 4] => "foo", "bar", "quack", "zap");
   ///
   /// assert_eq!(tv.swap_remove(1), "bar");
   /// assert_eq!(tv.as_slice(), &["foo", "zap", "quack"][..]);
@@ -639,6 +848,7 @@ pub struct TinyVecDrain<'p, A: Array> {
   target_index: usize,
   target_count: usize,
 }
+impl<'p, A: Array> FusedIterator for TinyVecDrain<'p, A> {}
 impl<'p, A: Array> Iterator for TinyVecDrain<'p, A> {
   type Item = A::Item;
   #[inline]
@@ -659,6 +869,113 @@ impl<'p, A: Array> Drop for TinyVecDrain<'p, A> {
   }
 }
 
+/// Splicing iterator for `TinyVec`
+/// See [`TinyVec::splice`](TinyVec::<A>::splice)
+pub struct TinyVecSplice<'p, A: Array, I: Iterator<Item = A::Item>> {
+  parent: &'p mut TinyVec<A>,
+  removal_start: usize,
+  removal_end: usize,
+  replacement: I,
+}
+
+impl<'p, A, I> Iterator for TinyVecSplice<'p, A, I>
+where
+  A: Array,
+  I: Iterator<Item = A::Item>,
+{
+  type Item = A::Item;
+
+  fn next(&mut self) -> Option<A::Item> {
+    if self.removal_start < self.removal_end {
+      match self.replacement.next() {
+        Some(replacement) => {
+          let removed = core::mem::replace(
+            &mut self.parent[self.removal_start],
+            replacement,
+          );
+          self.removal_start += 1;
+          Some(removed)
+        }
+        None => {
+          let removed = self.parent.remove(self.removal_start);
+          self.removal_end -= 1;
+          Some(removed)
+        }
+      }
+    } else {
+      None
+    }
+  }
+
+  #[inline]
+  fn size_hint(&self) -> (usize, Option<usize>) {
+    let len = self.len();
+    (len, Some(len))
+  }
+}
+
+impl<'p, A, I> ExactSizeIterator for TinyVecSplice<'p, A, I>
+where
+  A: Array,
+  I: Iterator<Item = A::Item>,
+{
+  #[inline]
+  fn len(&self) -> usize {
+    self.removal_end - self.removal_start
+  }
+}
+
+impl<'p, A, I> FusedIterator for TinyVecSplice<'p, A, I>
+where
+  A: Array,
+  I: Iterator<Item = A::Item>,
+{
+}
+
+impl<'p, A, I> DoubleEndedIterator for TinyVecSplice<'p, A, I>
+where
+  A: Array,
+  I: Iterator<Item = A::Item> + DoubleEndedIterator,
+{
+  fn next_back(&mut self) -> Option<A::Item> {
+    if self.removal_start < self.removal_end {
+      match self.replacement.next_back() {
+        Some(replacement) => {
+          let removed = core::mem::replace(
+            &mut self.parent[self.removal_end - 1],
+            replacement,
+          );
+          self.removal_end -= 1;
+          Some(removed)
+        }
+        None => {
+          let removed = self.parent.remove(self.removal_end - 1);
+          self.removal_end -= 1;
+          Some(removed)
+        }
+      }
+    } else {
+      None
+    }
+  }
+}
+
+impl<'p, A: Array, I: Iterator<Item = A::Item>> Drop
+  for TinyVecSplice<'p, A, I>
+{
+  fn drop(&mut self) {
+    for _ in self.by_ref() {}
+
+    let (lower_bound, _) = self.replacement.size_hint();
+    self.parent.reserve(lower_bound);
+
+    for replacement in self.replacement.by_ref() {
+      self.parent.insert(self.removal_end, replacement);
+      self.removal_end += 1;
+    }
+  }
+}
+
 impl<A: Array> AsMut<[A::Item]> for TinyVec<A> {
   #[inline(always)]
   #[must_use]
@@ -694,9 +1011,27 @@ impl<A: Array> BorrowMut<[A::Item]> for TinyVec<A> {
 impl<A: Array> Extend<A::Item> for TinyVec<A> {
   #[inline]
   fn extend<T: IntoIterator<Item = A::Item>>(&mut self, iter: T) {
-    for t in iter {
-      self.push(t)
-    }
+    let iter = iter.into_iter();
+    let (lower_bound, _) = iter.size_hint();
+    self.reserve(lower_bound);
+
+    let a = match self {
+      TinyVec::Heap(h) => return h.extend(iter),
+      TinyVec::Inline(a) => a,
+    };
+
+    let mut iter = a.fill(iter);
+    let maybe = iter.next();
+
+    let surely = match maybe {
+      Some(x) => x,
+      None => return,
+    };
+
+    let mut v = a.drain_to_vec_and_reserve(a.len());
+    v.push(surely);
+    v.extend(iter);
+    *self = TinyVec::Heap(v);
   }
 }
 
@@ -708,6 +1043,12 @@ impl<A: Array> From<ArrayVec<A>> for TinyVec<A> {
   }
 }
 
+impl<A: Array> From<A> for TinyVec<A> {
+  fn from(array: A) -> Self {
+    TinyVec::Inline(ArrayVec::from(array))
+  }
+}
+
 impl<T, A> From<&'_ [T]> for TinyVec<A>
 where
   T: Clone + Default,
@@ -744,9 +1085,7 @@ impl<A: Array + Default> FromIterator<A::Item> for TinyVec<A> {
   #[must_use]
   fn from_iter<T: IntoIterator<Item = A::Item>>(iter: T) -> Self {
     let mut av = Self::default();
-    for i in iter {
-      av.push(i)
-    }
+    av.extend(iter);
     av
   }
 }
@@ -770,7 +1109,7 @@ impl<A: Array> TinyVecIterator<A> {
     }
   }
 }
-
+impl<A: Array> FusedIterator for TinyVecIterator<A> {}
 impl<A: Array> Iterator for TinyVecIterator<A> {
   type Item = A::Item;
   #[inline]
@@ -811,7 +1150,10 @@ impl<A: Array> Iterator for TinyVecIterator<A> {
   }
 }
 
-impl<A: Array> Debug for TinyVecIterator<A> where A::Item: Debug {
+impl<A: Array> Debug for TinyVecIterator<A>
+where
+  A::Item: Debug,
+{
   #[allow(clippy::missing_inline_in_public_items)]
   fn fmt(&self, f: &mut Formatter<'_>) -> core::fmt::Result {
     f.debug_tuple("TinyVecIterator").field(&self.as_slice()).finish()
@@ -833,7 +1175,7 @@ impl<A: Array> IntoIterator for TinyVec<A> {
 
 impl<'a, A: Array> IntoIterator for &'a mut TinyVec<A> {
   type Item = &'a mut A::Item;
-  type IntoIter = alloc::slice::IterMut<'a, A::Item>;
+  type IntoIter = core::slice::IterMut<'a, A::Item>;
   #[inline(always)]
   #[must_use]
   fn into_iter(self) -> Self::IntoIter {
@@ -843,7 +1185,7 @@ impl<'a, A: Array> IntoIterator for &'a mut TinyVec<A> {
 
 impl<'a, A: Array> IntoIterator for &'a TinyVec<A> {
   type Item = &'a A::Item;
-  type IntoIter = alloc::slice::Iter<'a, A::Item>;
+  type IntoIter = core::slice::Iter<'a, A::Item>;
   #[inline(always)]
   #[must_use]
   fn into_iter(self) -> Self::IntoIter {
